# Copyright 2020, Ryan P. Kelly.

import argparse
import os
import shutil
import sys

from locald.client import Client
from locald.config import get_config, get_config_for_service
from locald.server import ensure_server, is_server_running, stop_server


class App(object):

    def run(self):

        parser = argparse.ArgumentParser()

        parser.add_argument(
            "--config",
            "-c",
            help="path to locald configuration file",
        )

        subparsers = parser.add_subparsers(dest="command")

        server_start_parser = subparsers.add_parser("server-start")
        server_start_parser.set_defaults(func=self.server_start)

        server_start_parser.add_argument(
            "--no-daemonize",
            action="store_true",
        )

        server_stop_parser = subparsers.add_parser("server-stop")
        server_stop_parser.set_defaults(func=self.server_stop)

        server_status_parser = subparsers.add_parser("server-status")
        server_status_parser.set_defaults(func=self.server_status)

        start_parser = subparsers.add_parser("start")
        start_parser.set_defaults(func=self.start)

        start_parser.add_argument("name")

        stop_parser = subparsers.add_parser("stop")
        stop_parser.set_defaults(func=self.stop)

        stop_parser.add_argument("name")

        status_parser = subparsers.add_parser("status")
        status_parser.set_defaults(func=self.status)

        status_parser.add_argument("names")

        logs_parser = subparsers.add_parser("logs")
        logs_parser.set_defaults(func=self.logs)

        logs_parser.add_argument("names")

        args = parser.parse_args()

        if not args.command:
            parser.print_help()
            sys.exit(0)

        config = get_config(args.config)

        returncode = args.func(config, args)

        return returncode

    def start(self, config, args):
        client = Client(config)
        client.start(args.name)

    def stop(self, config, args):
        client = Client(config)
        client.stop(args.name)

    def status(self, config, args):
        client = Client(config)
        client.status(args.names)

    def server_start(self, config, args):
        ensure_server(config, args)

    def server_stop(self, config, args):
        stop_server(config)

    def server_status(self, config, args):
        if is_server_running(config):
            sys.stdout.write("daemon is running\n")
            sys.stdout.flush()
            sys.exit(0)
        else:
            sys.stdout.write("daemon NOT running\n")
            sys.stdout.flush()
            sys.exit(1)

    def logs(self, config, args):

        names = [n.strip() for n in args.names.split(",") if n.strip()]

        if "ALL" in names:
            for key, values in config.items():
                if "service_path" in values:
                    names.append(key)

            names.remove("ALL")
            names = list(set(names))

        log_paths = []
        for name in names:
            service_config = get_config_for_service(config, name)

            if "log" not in service_config["service"]:
                continue

            log_paths.append(service_config["service"]["log"])

        if not log_paths:
            sys.stderr.write("NO logs to tail\n")
            sys.stderr.flush()
            sys.exit(1)

        tail_args = [
            "tail",
            "-F",
        ]

        tail_args.extend(log_paths)

        exec_path = shutil.which(tail_args[0])

        os.execv(exec_path, tail_args)



import json
import logging
import logging.config
import os
import queue
import select
import socket
import traceback

from daemonize import Daemonize

from .config import get_config, get_config_for_service
from .service import Service


logger = logging.getLogger()


class Socket(object):

    def __init__(self, path):
        self.path = path
        self.socket = None

    def __enter__(self):
        self.create()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.remove()

    def create(self):
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        try:
            os.unlink(self.path)
        except OSError:
            if os.path.exists(self.path):
                raise

        self.socket.bind(self.path)
        self.socket.listen()

    def remove(self):
        try:
            self.socket.close()
        except OSError:
            pass

        try:
            os.unlink(self.path)
        except OSError:
            pass


class Server(object):

    def __init__(self, config):
        self.config = config
        self.processes = {}

    def start(self):

        logging.config.fileConfig(self.config["locald"]["config_path"])

        try:
            return self._run()
        except:
            logger.error(traceback.format_exc())
            raise
        finally:
            for proc in self.processes.values():
                proc.kill()

    def _run(self):

        if "working_dir" in self.config["locald"]:
            os.chdir(self.config["locald"]["working_dir"])
        else:
            os.chdir(self.config["locald"]["config_dir"])

        socket_path = self.config["locald"]["socket_path"]

        logger.info(
            "[locald] going to begin listening on socket at {}"
            .format(socket_path)
        )

        timeout = 1  # 1 second

        with Socket(socket_path) as sock:

            inputs = [sock.socket]
            outputs = []
            messages_queue = {}

            while inputs:

                readable, writable, exceptional = select.select(
                    inputs,
                    outputs,
                    inputs,
                    timeout,
                )

                for s in readable:
                    if s is sock.socket:
                        connection, client_address = s.accept()
                        logger.info(
                            "[locald] new connection from {}"
                            .format(client_address)
                        )

                        connection.setblocking(0)
                        inputs.append(connection)

                        messages_queue[connection] = queue.Queue()
                    else:
                        data = s.recv(1024)
                        if data:
                            logger.debug(
                                "[locald] received '{}' from {}"
                                .format(data, s.getpeername())
                            )
                            messages_queue[s].put(data)

                            if s not in outputs:
                                outputs.append(s)
                        else:
                            logger.info(
                                "[locald] closing {} after reading no data"
                                .format(s.getpeername())
                            )

                            if s in outputs:
                                outputs.remove(s)

                            inputs.remove(s)
                            s.close()

                            del messages_queue[s]

                for s in writable:

                    response = None

                    try:
                        next_message = messages_queue[s].get_nowait()
                        response = self.process_message(next_message)
                    except KeyError:
                        logger.warning(
                            "[locald] output queue for {} is missing"
                            .format(s)
                        )
                    except queue.Empty:
                        logger.info(
                            "[locald] output queue for {} is empty"
                            .format(s.getpeername())
                        )
                        outputs.remove(s)
                    else:
                        logger.debug(
                            "[locald] sending '{}' to {}"
                            .format(response, s.getpeername())
                        )
                        s.send(response)

                for s in exceptional:
                    logger.info(
                        "[locald] handling exceptional condition for {}"
                        .format(s.getpeername())
                    )
                    inputs.remove(s)
                    if s in outputs:
                        outputs.remove(s)
                    s.close()

                    del messages_queue[s]

                self.tend_processes()

    def process_message(self, message):

        message = message.decode("utf-8")
        data = json.loads(message)

        if "command" in data:
            command = data["command"]

            if command == "start":
                response = self.handle_start(data)
            elif command == "stop":
                response = self.handle_stop(data)
            elif command == "status":
                response = self.handle_status(data)
            else:
                response = self.handle_unknown(data)
        else:
            response = self.handle_unknown(data)

        response = json.dumps(response).encode("utf-8")
        return response

    def handle_start(self, command):

        name = command["name"]

        if name not in self.config:
            return {
                "messages": ["unknown service '{}'".format(name)],
            }

        messages, _ = self.start_service(name)

        return {
            "messages": messages,
        }

    def start_service(self, name):

        service_config = get_config_for_service(self.config, name)

        requires = []
        if "requires" in service_config["service"]:
            requires = service_config["service"]["requires"]
            requires = [r.strip() for r in requires.split(",") if r.strip()]

        for require in requires:
            if require not in self.config:
                return ["unknown required service '{}'".format(require)], True

        messages = []

        for require in requires:
            sub_messages, is_error = self.start_service(require)
            messages.extend(sub_messages)
            if is_error:
                return messages, True

        if name not in self.processes:
            proc = Service(name, service_config)
            self.processes[name] = proc

        self.processes[name].start()

        messages.append("started '{}'".format(name))

        return messages, False

    def handle_stop(self, command):

        name = command["name"]

        if name not in self.config:
            return {
                "messages": ["unknown service '{}'".format(name)],
            }

        if name not in self.processes:
            message = "'{}' is not running".format(name)
        else:
            self.processes[name].kill()

            message = "kill signal sent to '{}'".format(name)

        return {
            "messages": [message],
        }

    def handle_status(self, command):

        name = command["name"]

        if name not in self.config:
            return {
                "messages": ["unknown service '{}'".format(name)],
            }

        if name in self.processes:
            if self.processes[name].is_running():
                status = "service {} is running".format(name)
            else:
                status = "service {} is NOT running".format(name)
        else:
            status = "service {} is NOT running".format(name)

        return {
            "messages": [status],
        }

    def handle_unknown(self, command):

        if "command" in command:
            message = "unknown command '{}'".format(command["command"])
        else:
            message = "invalid command '{}' received from client".format(command)

        return {
            "messages": [message],
        }

    def tend_processes(self):
        for proc in self.processes.values():
            proc.tend()


def get_pid(pid_path):

    with open(pid_path, "rt") as fp:
        pid = fp.read()

    pid = int(pid)

    return pid


def is_server_running(config):

    if not os.path.exists(config["locald"]["pid_path"]):
        return False

    pid = get_pid(config["locald"]["pid_path"])

    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


def create_server(config):
    server = Server(config)
    return server


def create_daemon(config, server):

    pid_path = config["locald"]["pid_path"]

    daemon = Daemonize(
        app="locald",
        pid=pid_path,
        action=server.start,
    )

    return daemon


def ensure_server(config, args):
    if not is_server_running(config):
        server = create_server(config)

        if args.no_daemonize:
            server.start()
        else:
            daemon = create_daemon(config, server)
            daemon.start()


def stop_server(config):

    if not os.path.exists(config["locald"]["pid_path"]):
        return False

    pid = get_pid(config["locald"]["pid_path"])

    try:
        os.kill(pid, 2)
    except OSError:
        return False
    else:
        return True



import datetime
import logging
import shlex
import subprocess


logger = logging.getLogger()


class Service(object):

    def __init__(self, name, config):
        self.name = name
        self.config = config
        self.process = None
        self.dead_since = None
        self.was_killed = False

    def tend(self):
        returncode = self.get_returncode()
        if returncode is not None:
            logger.info(
                "[locald] service {} pid {} exited with {}"
                .format(self.name, self.process.pid, returncode)
            )

            if self.process is not None:

                file_handles = [
                    self.process.stdout,
                    self.process.stderr,
                ]

                for file_handle in file_handles:
                    if hasattr(file_handle, "close"):
                        try:
                            file_handle.close()
                        except:
                            pass

            self.process = None
            if not self.was_killed:
                self.dead_since = datetime.datetime.now()
        else:
            if self.dead_since:
                restart = self.config["service"].get("restart", "never")
                if restart == "always":

                    restart_seconds = int(self.config["service"].get("restart_seconds", "0"))

                    restart_time = (
                        self.dead_since
                        + datetime.timedelta(seconds=restart_seconds)
                    )

                    if datetime.datetime.now() >= restart_time:
                        self.start()

        return returncode

    def start(self):
        if self.process is not None:
            logger.info(
                "[locald] service {} is already running, not starting"
                .format(self.name)
            )
            return

        logger.info(
            "[locald] going to start service {}"
            .format(self.name)
        )

        popen_args = {}
        if self.config["service"].get("log"):
            log_fp = open(self.config["service"]["log"], "ab+", buffering=0)
            popen_args = {
                "stdout": log_fp,
                "stderr": subprocess.STDOUT,
            }

        #if self.name == "backend":
        #    raise Exception(popen_args)

        args = shlex.split(self.config["service"]["command"])
        self.process = subprocess.Popen(
            args,
            **popen_args,
        )
        self.dead_since = None
        self.was_killed = False

    def get_returncode(self):
        if self.process is None:
            return None

        return self.process.poll()

    def kill(self):
        if self.process is None:
            return

        self.process.kill()
        self.was_killed = True

    def is_running(self):
        if self.process is None:
            return False

        return self.get_returncode() is None



# Copyright 2017-2019, Ryan P. Kelly.

from setuptools import setup


setup(
    name="locald",
    version="0.3",
    description="run local services/microservices for development",
    author="Ryan P. Kelly",
    author_email="ryan@ryankelly.us",
    url="https://github.com/f0rk/locald",
    install_requires=[
        "daemonize",
    ],
    tests_require=[
        "pytest",
    ],
    package_dir={"": "lib"},
    packages=["locald"],
    scripts=["tools/locald"],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development",
    ],
)



import json
import socket

from .server import is_server_running


class Client(object):

    def __init__(self, config):
        self.config = config

    def connect(self):

        socket_path = self.config["locald"]["socket_path"]

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(socket_path)

        return sock

    def prepare_data(self, data):

        raw_data = json.dumps(data)
        raw_data = raw_data.encode("utf-8")

        return raw_data

    def parse_response(self, raw_data):
        data = raw_data.decode("utf-8")
        data = json.loads(data)

        return data

    def send(self, sock, command):

        data = self.prepare_data(command)
        sock.sendall(data)

        raw_response = sock.recv(1024)  # XXX

        return self.parse_response(raw_response)

    def send_command(self, command):

        sock = None

        try:
            sock = self.connect()
            response = self.send(sock, command)
        except FileNotFoundError:
            if is_server_running(self.config):
                message = "sending command failed. are your socket permissions correct?"
            else:
                message = "sending command failed. server does not appear to be running."

            response = {
                "messages": [message],
            }

        finally:
            if sock is not None:
                sock.close()

        return response

    def start(self, name):

        command = {
            "command": "start",
            "name": name,
        }

        response = self.send_command(command)

        for message in response["messages"]:
            print(message)

    def stop(self, name):

        command = {
            "command": "stop",
            "name": name,
        }

        response = self.send_command(command)

        for message in response["messages"]:
            print(message)

    def status(self, names):

        command = {
            "command": "status",
            "name": names,
        }

        response = self.send_command(command)

        for message in response["messages"]:
            print(message)



import configparser
import copy
import os.path


class ConfigError(ValueError):
    pass


def get_config_path(config_path):

    if config_path is not None:
        return os.path.realpath(config_path)

    maybe_config_dir = os.path.realpath(".")

    while True:

        maybe_config_path = os.path.join(
            maybe_config_dir,
            "locald.ini",
        )

        if os.path.exists(maybe_config_path):
            return maybe_config_path

        if maybe_config_dir == "/":
            raise Exception("Unable to locate configuration file")

        maybe_config_dir, _ = os.path.split(maybe_config_dir)


def get_config(config_path):

    config_path = get_config_path(config_path)
    config_dir, _ = os.path.split(config_path)

    parser = configparser.ConfigParser()
    parser.read(config_path)

    config = copy.deepcopy(parser._sections)

    if "locald" in config:
        key = "locald"
    elif "service" in config:
        key = "service"
    else:
        raise ConfigError(
            "incorrectly structured config file '{}' {}"
            .format(config_path, config)
        )

    config[key]["config_path"] = config_path
    config[key]["config_dir"] = config_dir

    return config


def get_config_for_service(config, name):

    service_config_path = config[name]["service_path"]

    # relative paths are resolved based on the base config's location
    if not os.path.isabs(service_config_path):
        config_dir = config["locald"]["config_dir"]
        service_config_path = os.path.join(config_dir, service_config_path)

    service_config = get_config(service_config_path)

    if "log" in service_config:
        log_path = service_config["log"]
        if  not os.path.isabs(log_path):
            config_dir = config["locald"]["config_dir"]
            log_path = os.path.join(config_dir, log_path)

        service_config["log"] = log_path

    return service_config
